package cz.cuni.dsrg.sofa.netbeans.visualdesigner.designer;

import javax.swing.*;
import javax.swing.event.MouseInputListener;
import java.awt.*;
import java.awt.event.MouseEvent;

/**
 * This class implements a JPanel, a layer that handles the incomming
 * mouse event. It is able to select the component and architectures,
 * change the visual size and location, and design the interconnections.
 *
 * @author Michael Cifka
 */
public class EventLayer extends JPanel implements MouseInputListener
{

    static final int SELECTION_MODE = 0;
    static final int INSTANCE_MODE = 10;
    static final int INTERFACE_MODE = 11;
    static final int BINDING_MODE = 12;

    private static final int GRID = Constants.GRID_SIZE;
    private static final int MOVE_MODE = 9;

    private ArchitectureEditor editor;
    private JComponent selection;
    private VisualDesign visualDesign;
    private Rectangle selFrame;
    private Rectangle dragFrame;
    private Rectangle lastFrame;
    private Point dragOrigin;
    private int mode;
    private JComponent ghostComponent;

    private final int[] cursors = new int[]{
        Cursor.DEFAULT_CURSOR, Cursor.N_RESIZE_CURSOR,
        Cursor.NE_RESIZE_CURSOR, Cursor.E_RESIZE_CURSOR,
        Cursor.SE_RESIZE_CURSOR, Cursor.S_RESIZE_CURSOR,
        Cursor.SW_RESIZE_CURSOR, Cursor.W_RESIZE_CURSOR,
        Cursor.NW_RESIZE_CURSOR, Cursor.MOVE_CURSOR,
        Cursor.HAND_CURSOR, Cursor.HAND_CURSOR,
        Cursor.CROSSHAIR_CURSOR};

    public EventLayer(ArchitectureEditor editor)
    {
        setOpaque(false);
        this.editor = editor;
        addMouseListener(this);
        addMouseMotionListener(this);
    }

    public void mouseClicked(MouseEvent mouseEvent)
    {
        // handle double-clicks
        if(mouseEvent.getClickCount() == 2)
        {
            JComponent jc = editor.getJComponentAt(mouseEvent.getPoint());
            if(jc instanceof GrInstance)
                editor.getArchitectureWindow().moveToInstance(((GrInstance) jc).getInstance());
            else if(jc instanceof GrFrame)
            {
                editor.getArchitectureWindow().moveToParent();
            }
        }
    }

    public void mouseDragged(MouseEvent mouseEvent)
    {
        if(mode == 0) return;

        int x = ((int) ((mouseEvent.getX() - dragOrigin.x) / GRID)) * GRID;
        int y = ((int) ((mouseEvent.getY() - dragOrigin.y) / GRID)) * GRID;

        Rectangle rect = new Rectangle(selFrame);

        if(lastFrame == null)
            lastFrame = selection.getBounds();

        if(mode == MOVE_MODE)
        {
            rect.translate(x, y);
            if(visualDesign != null)
                rect = visualDesign.reviewMove(lastFrame, rect);
            setDragFrame(rect);
            return;
        }

        boolean north = (mode == SwingConstants.NORTH
                || mode == SwingConstants.NORTH_EAST
                || mode == SwingConstants.NORTH_WEST);
        boolean south = (mode == SwingConstants.SOUTH
                || mode == SwingConstants.SOUTH_EAST
                || mode == SwingConstants.SOUTH_WEST);
        boolean east = (mode == SwingConstants.EAST
                || mode == SwingConstants.SOUTH_EAST
                || mode == SwingConstants.NORTH_EAST);
        boolean west = (mode == SwingConstants.WEST
                || mode == SwingConstants.SOUTH_WEST
                || mode == SwingConstants.NORTH_WEST);

        if(west)
        {
            if(x > rect.width) x = rect.width;
            rect.x += x;
            rect.width -= x;
        }
        if(east)
        {
            if(x < -rect.width) x = -rect.width;
            rect.width += x;
        }
        if(north)
        {
            if(y > rect.height) y = rect.height;
            rect.y += y;
            rect.height -= y;
        }
        if(south)
        {
            if(y < -rect.height) y = -rect.height;
            rect.height += y;
        }

        // respect minimum size
        Dimension minSize = selection.getMinimumSize();
        if(rect.width < minSize.width)
        {
            if(west) rect.x -= minSize.width - rect.width;
            rect.width = minSize.width;
        }
        if(rect.height < minSize.height)
        {
            if(north) rect.y -= minSize.height - rect.height;
            rect.height = minSize.height;
        }

        if(visualDesign != null)
            rect = visualDesign.reviewResize(lastFrame, rect);
        setDragFrame(rect);
    }

    public void mouseEntered(MouseEvent mouseEvent)
    {
        ToolTipManager.sharedInstance().mouseEntered(mouseEvent);
    }

    public void mouseExited(MouseEvent mouseEvent)
    {
        ToolTipManager.sharedInstance().mouseExited(mouseEvent);
    }

    public void mouseMoved(MouseEvent mouseEvent)
    {
        if(mode == INSTANCE_MODE)
        {
            mouseMoved2(mouseEvent);
            return;
        }

        if(mode == INTERFACE_MODE)
        {
            mouseMoved3(mouseEvent);
            return;
        }

        if(mode == BINDING_MODE)
        {
            mouseMoved4(mouseEvent);
            return;
        }

        ToolTipManager.sharedInstance().mouseMoved(mouseEvent);

        selection = editor.getSelection();
        JComponent c = editor.getJComponentAt(mouseEvent.getPoint());
        if(selection == null || selection != c)
        {
            setMode(0);
            return;
        }

        selFrame = selection.getBounds();
        visualDesign = (selection instanceof VisualDesign)
                ? (VisualDesign) selection : null;
        dragOrigin = mouseEvent.getPoint();


        int x = mouseEvent.getX(), y = mouseEvent.getY();

        // check if we are resizing
        final int t = Constants.EDGE_TOLERANCE;
        Rectangle exSelFrame = new Rectangle(selFrame);
        exSelFrame.grow(t, t);
        if(exSelFrame.contains(x, y))
        {
            boolean north = selFrame.y - t <= y && y <= selFrame.y + t;
            boolean west = selFrame.x - t <= x && x <= selFrame.x + t;
            boolean south = selFrame.y + selFrame.height - t <= y
                    && y <= selFrame.y + selFrame.height + t;
            boolean east = selFrame.x + selFrame.width - t <= x
                    && x <= selFrame.x + selFrame.width + t;

            int direction = 0;
            if(south && east)
                direction = SwingConstants.SOUTH_EAST;
            else if(north && east)
                direction = SwingConstants.NORTH_EAST;
            else if(south && west)
                direction = SwingConstants.SOUTH_WEST;
            else if(north && west)
                direction = SwingConstants.NORTH_WEST;
            else if(east)
                direction = SwingConstants.EAST;
            else if(south)
                direction = SwingConstants.SOUTH;
            else if(west)
                direction = SwingConstants.WEST;
            else if(north) direction = SwingConstants.NORTH;

            if(direction != 0)
                if(visualDesign == null || visualDesign.canResize(direction))
                {
                    setMode(direction);
                    return;
                }
        }

        // check if we are moving
        if(selFrame.contains(x, y))
            if(visualDesign == null || visualDesign.canMove())
            {
                setMode(MOVE_MODE);
                return;
            }

        // if not either set mode to zero
        setMode(0);
    }

    // handler for the instance mode
    public void mouseMoved2(MouseEvent mouseEvent)
    {
        if(editor.getJComponentAt(mouseEvent.getPoint()) != null)
        {
            Point pt = alignToGrid(mouseEvent.getPoint());
            setDragFrame(((VisualDesign) ghostComponent).reviewMove(null,
                                                                    new Rectangle(pt, ghostComponent.getSize())));
        }
        else
            setDragFrame(null);
    }

    // handler for the interface mode
    public void mouseMoved3(MouseEvent mouseEvent)
    {
        Point p = mouseEvent.getPoint();
        Point tp = new Point(p);
        JComponent c = getFrameEdge(p, tp);
        final int t = Constants.EDGE_TOLERANCE + Constants.INTERFACE_SIZE / 2 + 1;
        if(c == null)
        {
            tp.translate(-t, -t);
            c = getFrameEdge(p, tp);
        }
        if(c == null)
        {
            tp.translate(2 * t, 0);
            c = getFrameEdge(p, tp);
        }
        if(c == null)
        {
            tp.translate(0, 2 * t);
            c = getFrameEdge(p, tp);
        }
        if(c == null)
        {
            tp.translate(-2 * t, 0);
            c = getFrameEdge(p, tp);
        }

        if(c == null)
        {
            editor.setSelection(null);
            setDragFrame(null);
            return;
        }
        p.translate(Constants.EDGE_TOLERANCE, Constants.EDGE_TOLERANCE);
        p = alignToGrid(p);

        // exclude corner positions
        int minx = c.getX(), maxx = c.getX() + c.getWidth() - 1;
        int miny = c.getY(), maxy = c.getY() + c.getHeight() - 1;
        if((p.x == minx || p.x == maxx) && (p.y == miny || p.y == maxy))
        {
            editor.setSelection(null);
            setDragFrame(null);
            return;
        }

        editor.setSelection(c);

        final int interfaceSize = Constants.INTERFACE_SIZE;
        p.translate(-interfaceSize / 2, -interfaceSize / 2);
        setDragFrame(new Rectangle(p.x, p.y, interfaceSize, interfaceSize));
    }

    // handler for binding mode
    public void mouseMoved4(MouseEvent mouseEvent)
    {
        JComponent c = editor.getJComponentAt(mouseEvent.getPoint());
        if(c instanceof GrInterface)
            setDragFrame(c.getBounds());
        else
            setDragFrame(null);
        dragOrigin = mouseEvent.getPoint();
    }

    private JComponent getFrameEdge(Point point, Point testPoint)
    {
        JComponent c = editor.getJComponentAt(testPoint);
        if(c == null || !(c instanceof GrFrame || c instanceof GrInstance))
            return null;
        Rectangle r = new Rectangle(c.getBounds());
        final int t = Constants.EDGE_TOLERANCE;
        r.grow(t, t);
        if(!r.contains(point)) return null;
        r.grow(-2 * t - 1, -2 * t - 1);
        if(r.contains(point)) return null;
        return c;
    }


    public void mousePressed(MouseEvent mouseEvent)
    {
        if(mode == SELECTION_MODE)
        {
            editor.setSelection(editor.getJComponentAt(mouseEvent.getPoint()));
            repaint();
            mouseMoved(mouseEvent);
        }
        else if(mode == INSTANCE_MODE || mode == INTERFACE_MODE)
        {
            if(dragFrame != null)
            {
                if(mode == INSTANCE_MODE) editor.addInstance(dragFrame);
                if(mode == INTERFACE_MODE)
                {
                    JComponent selection = editor.getSelection();
                    if(selection != null)
                    {
                        Point center = dragFrame.getLocation();
                        center.translate(Constants.INTERFACE_SIZE / 2,
                                         Constants.INTERFACE_SIZE / 2);
                        editor.addInterface(center, selection);
                    }
                }
                setDragFrame(null);
            }
            setMode(SELECTION_MODE);
        }
        else if(mode == BINDING_MODE)
        {
            JComponent c = editor.getJComponentAt(mouseEvent.getPoint());
            if(c instanceof GrInterface)
            {
                if(selection instanceof GrInterface
                        && ((GrInterface) selection).canBind((GrInterface) c))
                {
                    // bind selection to current point
                    editor.addBinding((GrInterface) selection, (GrInterface) c);
                    editor.setSelection(null);
                    setDragFrame(null);
                    setMode(SELECTION_MODE);
                }
                else
                {
                    editor.setSelection(c);
                    repaint();
                }
            }
        }
    }

    public void mouseReleased(MouseEvent mouseEvent)
    {
        if(dragFrame != null && mode != BINDING_MODE)
        {
            selection.setBounds(dragFrame);
            selection.validate();
            setDragFrame(null);
            mouseMoved(mouseEvent);
        }
    }

    public String getToolTipText(MouseEvent mouseEvent)
    {

        Component c = editor.getJComponentAt(mouseEvent.getPoint());
        return (c != null) ? ((JComponent) c).getToolTipText() : null;
    }

    void setMode(int mode)
    {
        this.mode = mode;
        setCursor(Cursor.getPredefinedCursor(cursors[mode]));

        if(mode == INSTANCE_MODE)
            ghostComponent = new GrInstance(editor.getFrame());
        if(mode == BINDING_MODE)
            dragOrigin = new Point(0, 0);

        repaint();
    }


    protected void paintComponent(Graphics g)
    {
        selection = editor.getSelection();
        if(selection != null)
        {
            selFrame = selection.getBounds();
            g.setColor(Color.blue);
            g.drawRect(selFrame.x, selFrame.y, selFrame.width - 1, selFrame.height - 1);
            if(mode == BINDING_MODE)
            {
                JComponent comp = editor.getJComponentAt(dragOrigin);
                if((comp instanceof GrInterface) && (selection instanceof GrInterface))
                {
                    GrInterface ifaceSel = (GrInterface) selection;
                    GrInterface ifaceCur = (GrInterface) comp;
                    if(ifaceSel.canBind(ifaceCur))
                    {
                        g.setColor(Color.red);
                        if(ifaceSel.isSource())
                            BindingLayer.drawBinding(g, ifaceSel, ifaceCur);
                        else
                            BindingLayer.drawBinding(g, ifaceCur, ifaceSel);
                    }
                }
            }
        }
        if(dragFrame != null)
        {
            g.setColor(Color.red);
            g.drawRect(dragFrame.x, dragFrame.y, dragFrame.width - 1, dragFrame.height - 1);
        }
    }


    private void setDragFrame(Rectangle r)
    {
        lastFrame = (r == null) ? null : dragFrame;
        dragFrame = r;
        repaint();
    }

    private Point alignToGrid(Point pt)
    {
        int x = pt.x - Constants.FRAME_INSETS;
        int y = pt.y - Constants.FRAME_INSETS;
        x = ((int) (x / GRID)) * GRID + Constants.FRAME_INSETS;
        y = ((int) (y / GRID)) * GRID + Constants.FRAME_INSETS;
        return new Point(x, y);
    }

    private Point getCenter(Rectangle rect)
    {
        return new Point(
                (int) (rect.getX() + rect.getWidth() / 2),
                (int) (rect.getY() + rect.getHeight() / 2)
        );
    }
}
