package cz.cuni.dsrg.sofa.netbeans.visualdesigner.objects;

import cz.cuni.dsrg.sofa.netbeans.visualdesigner.designer.Layout;
import cz.cuni.dsrg.sofa.netbeans.visualdesigner.types.*;

import java.beans.PropertyChangeListener;
import java.beans.PropertyChangeSupport;
import java.util.*;

/**
 * This class represents a software component.<p>
 *
 * Methods to set the name, type and implementation are provided as well
 * as methods to manage the set of interfaces and the set of properties.<p>
 *
 * The component is able to build up itself based on a provided type. The
 * internal structure of the component (the assembly tree) can be written
 * to a deployment descriptor (.dd) file recursivelly.<p>
 *
 * Values of the properties can be set and the class also provides a color
 * for colouring the deployment units, set of subcomponent that are set to be
 * deployed to the same server.<p>
 *
 * This class can be persisted using the Java Object Serialization.
 *
 * @author Michael Cifka
 */
public class Component implements java.io.Serializable, PropertiesProvider
{

    Component()
    {
    }

    private Component(String name, FrameType type, Repository repository)
    {
        this.name = name;
        this.repository = repository;
        this.parent = null;
        this.commonData = new CommonData();
        setType(type);
    }

    public Component(String name, FrameType type, Architecture parent)
    {
        this(name, type, parent.getParent().getRepository());
        this.parent = parent;
        this.commonData = null;
        createImplementation();
    }


    // constructor to be used only by the component loader
    public Component(String name, Repository repository)
    {
        this(name, repository.createFrameType(), repository);
        createImplementation();
    }

    public String getName()
    {
        return name;
    }

    public void setName(String name)
    {
        // if part of an architecture, perform through the architecture
        if(parent != null)
            parent.renameInstance(this, name);
        else
        {
            this.name = name;
            firePropertyChange("name");
        }
    }

    void rename(String name)
    {
        this.name = name;
    }

    public Repository getRepository()
    {
        return repository;
    }

    public FrameType getType()
    {
        return type;
    }

    public void setType(FrameType type)
    {

        if(type == this.type)
        {
            firePropertyChange("type");
            return;
        }

        // remove current bindings
        if(parent != null && interfaces != null)
        {
            Interface[] ifaces = getInterfaces();
            for(int i = 0; i < ifaces.length; i++)
            {
                Binding[] binds = ifaces[i].getBindings();
                for(int j = 0; j < binds.length; j++)
                    parent.removeBinding(binds[j]);
            }
        }

        // modify references
        if(this.type != null) removeReferences();
        if(type != null) type.addReference();

        this.type = type;
        implementation = null;
        interfaces = new ArrayList();
        interfacesIndex = new HashMap();
        propertyTypes = new TreeMap();
        propertyValues = new HashMap();
        arrayProperty = null;
        laidout = false;

        if(type != null)
        {
            FrameType.Interface[] provisions = type.getProvisions();
            for(int i = 0; i < provisions.length; i++)
            {
                Interface iface = new Interface(
                        provisions[i].getName(), provisions[i].getType(),
                        Interface.PROVIDED, this
                );
                if(provisions[i].getArray() != null)
                    iface.setArrayProperty(provisions[i].getArray().getProperty());
                interfacesIndex.put(iface.getName(), iface);
                interfaces.add(iface);
            }
            FrameType.Interface[] requirements = type.getRequirements();
            for(int i = 0; i < requirements.length; i++)
            {
                Interface iface = new Interface(
                        requirements[i].getName(), requirements[i].getType(),
                        Interface.REQUIRED, this
                );
                if(requirements[i].getArray() != null)
                    iface.setArrayProperty(requirements[i].getArray().getProperty());
                interfacesIndex.put(iface.getName(), iface);
                interfaces.add(iface);
            }
            Property[] properties = type.getProperties();
            for(int i = 0; i < properties.length; i++)
            {
                addProperty(properties[i].getName(), properties[i].getType());
            }
        }

        // modify the associated type
        if(parent != null)
            parent.getType().getInstance(name).setType(type);

        firePropertyChange("type");
    }

    public Architecture getParent()
    {
        return parent;
    }

    public Architecture getImplementation()
    {
        return implementation;
    }

    public void setImplementation(Architecture impl)
    {
        if(implementation != null)
            implementation.removeReferences();
        implementation = impl;
        firePropertyChange("implementation");
    }

    public Architecture createImplementation()
    {
        ArchitectureType implType = repository.createArchitectureType(getType());
        Architecture impl = new Architecture(implType, this);
        setImplementation(impl);
        return impl;
    }

    public Interface[] getInterfaces()
    {
        return (Interface[]) interfaces.toArray(new Interface[0]);
    }

    public Interface getInterface(String name)
    {
        return (Interface) interfacesIndex.get(name);
    }

    public Interface createInterface(int mode)
    {
        InterfaceType ifaceType = repository.createInterfaceType();
        Interface iface = new Interface(createInterfaceName(), ifaceType, mode, this);
        return iface;
    }

    public String createInterfaceName()
    {
        String base = "iface";
        int i = 1;
        String s = base + i++;
        while(getInterface(s) != null)
        {
            s = base + i++;
        }
        return s;
    }

    public void addInterface(Interface iface)
    {
// add interface to the collection
        interfacesIndex.put(iface.getName(), iface);
        interfaces.add(iface);
// modify the associated type object
        if(iface.getMode() == Interface.PROVIDED)
            type.addInterface(iface.getName(), iface.getType(), FrameType.PROVISION);
        else
            type.addInterface(iface.getName(), iface.getType(), FrameType.REQUIREMENT);
        firePropertyChange("interfaces");
    }

    public void removeInterface(Interface iface)
    {
// remove interface from the collection
        interfacesIndex.remove(iface.getName());
        interfaces.remove(iface);

// remove bindings
        Binding[] binds = iface.getBindings();
        for(int i = 0; i < binds.length; i++)
        {
            if(implementation != null) implementation.removeBinding(binds[i]);
            if(parent != null) parent.removeBinding(binds[i]);
        }

// modify the references
        iface.removeReferences();

// modify the type
        type.removeInterface(iface.getName());

        firePropertyChange("interfaces");
    }



    // PROPERTIES

    private Map propertyTypes;
    private Map propertyValues;

    public String[] getProperties()
    {
        return (String[]) propertyTypes.keySet().toArray(new String[0]);
    }

    public void addProperty(String name, String type)
    {
        propertyTypes.put(name, type);
        propertyValues.remove(name);
        getType().addProperty(name, type);
        firePropertyChange("properties");
    }

    public void removeProperty(String name)
    {
        propertyTypes.remove(name);
        propertyValues.remove(name);
        getType().removeProperty(name);
        firePropertyChange("properties");
    }

    public String getPropertyType(String name)
    {
        return (String) propertyTypes.get(name);
    }

    public void setPropertyType(String name, String type)
    {
        addProperty(name, type);
    }

    public Object getPropertyValue(String name)
    {
        Object value = propertyValues.get(name);
        if(value == null)
        {
            // DEFAULT VALUES
            String type = getPropertyType(name);
            if(type == null)
                return null;
            if(type.equals("string"))
                setPropertyValue(name, "");
            else if(type.equals("integer"))
                setPropertyValue(name, new Integer(0));
            else
                return null;
            value = getPropertyValue(name);
        }
        return value;
    }

    public String getPropertyValueAsString(String name)
    {
        Object value = getPropertyValue(name);
        if(value instanceof String) return (String) value;
        if(value instanceof Integer) return "" + ((Integer) value).intValue();
        return "";
    }

    public void setPropertyValue(String name, Object value)
    {
        propertyValues.put(name, value);
    }

    // ARRAYS

    private String arrayProperty;

    public void setArrayProperty(String property)
    {
        if(parent != null && property != arrayProperty
                && (property == null || !property.equals(arrayProperty)))
        {

            arrayProperty = property;
            if(property == null)
                parent.getType().getInstance(name).setArray(null);
            else
            {
                ArrayType arrayType = getType().getArrayType(property);
                parent.getType().getInstance(name).setArray(arrayType);
            }
            firePropertyChange("array");
        }
    }

    public String getArrayProperty()
    {
        return arrayProperty;
    }

    // REFERENCES

    public void removeReferences()
    {
        Interface[] ifaces = getInterfaces();
        for(int i = 0; i < ifaces.length; i++)
            ifaces[i].removeReferences();

        if(implementation != null)
            implementation.removeReferences();

        if(type != null)
            type.removeReference();

        removeLocation(location);
        location = null;
    }

    // SOURCE & LOCATION

    private String location = null;
    private String source = null;

    public String getLocation()
    {
        return location;
    }

    public void setLocation(String s)
    {
        removeLocation(location);
        location = s;
        addLocation(location);

        firePropertyChange("location");
    }

    public String getSource()
    {
        return source;
    }

    public void setSource(String s)
    {
        source = s;
    }


    // EVENT SOURCE

    public void addPropertyChangeListener(PropertyChangeListener listener)
    {
        support.addPropertyChangeListener(listener);
    }

    public void removePropertyChangeListener(PropertyChangeListener listener)
    {
        support.removePropertyChangeListener(listener);
    }

    public void firePropertyChange(String propName)
    {
        // fire property change on the support
        support.firePropertyChange(propName, null, null);

        // propagate the property change to the higher level
        if(getParent() != null)
            getParent().getParent().firePropertyChange(propName);
    }

    private void readObject(java.io.ObjectInputStream s)
            throws java.io.IOException, ClassNotFoundException
    {
        s.defaultReadObject();
        support = new PropertyChangeSupport(this);
    }

    public Layout getLayout()
    {
        return layout;
    }

    public boolean isLaidout()
    {
        return laidout;
    }

    public void setLaidout(boolean b)
    {
        laidout = b;
    }

    public String writeAsDeployment()
    {
        String s = "<deployment>\n";
        s += writeAsDeployment(1, 0);
        s += "</deployment>\n";
        return s;
    }

    private String writeAsDeployment(int indent, int index)
    {
        String indentString = indentString(indent);
        if(implementation == null)
            return indentString + "<!-- Error: component name=\""
                    + name + "\" is incomplete! -->\n";
        ArchitectureType archType = implementation.getType();
        Component[] comps = implementation.getInstances();
        String[] compProps = getProperties();
        String[] archProps = implementation.getProperties();
        String s = indentString + "<component name=\"" + name;
        if(index != 0)
            s += "[" + index + "]";
        s += "\" provider=\"" + archType.getProviderName() + "\"\n";
        s += indentString + " architecture=\""
                + archType.getRelativeName(null) + "\"";
        if(location != null || source != null) s += "\n";
        if(source != null) s += indentString + " source=\"" + source + "\"";
        if(location != null) s += indentString + " location=\"" + location + "\"";
        if(comps.length == 0 && compProps.length == 0 && archProps.length == 0)
            return s + "/>\n";
        s += ">\n";
        for(int i = 0; i < compProps.length; i++)
            s += indentString(indent + 1) + "<property name=\""
                    + compProps[i] + "\" value=\""
                    + getPropertyValueAsString(compProps[i]) + "\"/>\n";
        for(int i = 0; i < archProps.length; i++)
            s += indentString(indent + 1) + "<property name=\""
                    + archProps[i] + "\" value=\""
                    + implementation.getPropertyValueAsString(archProps[i]) + "\"/>\n";
        for(int i = 0; i < comps.length; i++)
        {
            String arrayProperty = comps[i].getArrayProperty();
            if(arrayProperty == null)
                s += comps[i].writeAsDeployment(indent + 1, 0);
            else
            {
                Integer value = (Integer) implementation.getPropertyValue(arrayProperty);
                if(value == null)
                    value = (Integer) getPropertyValue(arrayProperty);
                //Debug.print("Value of: " + arrayProperty + " is null");
                for(int j = 1; j <= value.intValue(); j++)
                    s += comps[i].writeAsDeployment(indent + 1, j);
            }
        }

        s += indentString + "</component>\n";
        return s;
    }

    private String indentString(int indent)
    {
        String s = "";
        for(int i = 0; i < indent; i++)
            s = s + "     ";
        return s;
    }

    // COMMON DATA

    private static final int COLORS = 6;

    private static class CommonData implements java.io.Serializable
    {
        // default containers
        public String defaultModule = null;
        public String defaultProvider = null;

        // mappings for deployment units coloring
        public Map mapLocationColor = new HashMap();
        public Map mapLocationCount = new HashMap();
        public int[] arrColorCount = new int[COLORS];
    }

    private CommonData commonData = null;

    public CommonData getCommonData()
    {
        if(commonData == null)
            commonData = parent.getParent().getCommonData();
        return commonData;
    }

    // DEPLOYMENT UNITS COLORING

    private static class IntegerValue implements java.io.Serializable
    {
        private int value;

        public IntegerValue(int value)
        {
            this.value = value;
        }

        public int getValue()
        {
            return value;
        }

        public void setValue(int value)
        {
            this.value = value;
        }

        public int increase()
        {
            return ++value;
        }

        public int decrease()
        {
            return --value;
        }
    }

    private void addLocation(String s)
    {
        if(s == null) return;
        CommonData common = getCommonData();
        IntegerValue cntLocation = (IntegerValue) common.mapLocationCount.get(s);
        if(cntLocation == null)
        {
            // find best color
            int color = 0, count, rv = 9999;
            for(int i = 0; i < COLORS; i++)
                if(common.arrColorCount[i] < rv)
                {
                    color = i;
                    rv = common.arrColorCount[i];
                }
            common.arrColorCount[color]++;

            // add new location
            common.mapLocationCount.put(s, new IntegerValue(1));
            common.mapLocationColor.put(s, new IntegerValue(color));

        }
        else
            cntLocation.increase();
    }

    private void removeLocation(String s)
    {
        if(s == null) return;
        CommonData common = getCommonData();
        IntegerValue cntLocation = (IntegerValue) common.mapLocationCount.get(s);
        if(cntLocation.decrease() == 0)
        {
            common.mapLocationCount.remove(s);
            IntegerValue color = (IntegerValue) common.mapLocationColor.remove(s);
            common.arrColorCount[color.getValue()]--;
        }
    }

    public int getColorIndex()
    {
        if(location == null && parent != null)
            return parent.getParent().getColorIndex();

        CommonData common = getCommonData();
        IntegerValue color = (IntegerValue) common.mapLocationColor.get(location);
        if(color == null)
            return 0;
        else
            return color.getValue() + 1;
    }

    private String name;
    private FrameType type;
    private Architecture parent;
    private Architecture implementation;
    private Collection interfaces = new ArrayList();
    private Map interfacesIndex;
    private Repository repository;
    private Layout layout = new Layout();
    private boolean laidout = false;

    private transient PropertyChangeSupport support = new PropertyChangeSupport(this);
}