package cz.cuni.dsrg.sofa.netbeans.visualdesigner.objects;

import cz.cuni.dsrg.sofa.netbeans.visualdesigner.designer.Layout;
import cz.cuni.dsrg.sofa.netbeans.visualdesigner.types.ArrayType;
import cz.cuni.dsrg.sofa.netbeans.visualdesigner.types.FrameType;
import cz.cuni.dsrg.sofa.netbeans.visualdesigner.types.InterfaceType;

import java.util.Collection;
import java.util.LinkedList;

/**
 * This class represents an interface used in a software component.<p>
 *
 * Methods to set the name, type, mode, and to manage the set of bindings
 * are provided.<p>
 *
 * This class can be persisted using the Java Object Serialization.
 *
 * @author Michael Cifka
 */
public class Interface implements java.io.Serializable
{

    public static final int PROVIDED = 0;
    public static final int REQUIRED = 1;

    public Interface(String name, InterfaceType type, int mode, Component parent)
    {
        this.name = name;
        this.type = type;
        this.mode = mode;
        this.parent = parent;
        bindings = new LinkedList();
        arrayProperty = null;
        type.addReference();
    }


    public String getName()
    {
        return name;
    }

    public void setName(String str)
    {
        String old = name;
        // set the new name
        name = str;

        // modify the frame
        getParent().getType().renameInterface(old, str);


        getParent().firePropertyChange("interfaceName");
    }

    public InterfaceType getType()
    {
        return type;
    }

    public void setType(InterfaceType type)
    {
        if(this.type != type)
        {
            // first remove all bindings
            Binding[] binds = getBindings();
            for(int i = 0; i < binds.length; i++)
                binds[i].getParent().removeBinding(binds[i]);

            // modify references
            removeReferences();
            type.addReference();

            // then change the type
            this.type = type;

            //modify the frame
            getParent().getType().getInterface(name).setType(type);

        }
        getParent().firePropertyChange("interfaceType");
    }

    public int getMode()
    {
        return mode;
    }

    public void setMode(int mode)
    {
        if(this.mode == mode) return;

        // first remove all bindings
        Binding[] binds = getBindings();
        for(int i = 0; i < binds.length; i++)
            binds[i].getParent().removeBinding(binds[i]);

        this.mode = mode;

        // modify the frame
        getParent().getType().getInterface(name).setMode(
                (mode == PROVIDED) ? FrameType.PROVISION : FrameType.REQUIREMENT
        );

        getParent().firePropertyChange("interfaceMode");
    }

    public Component getParent()
    {
        return parent;
    }

    public Binding[] getBindings()
    {
        return (Binding[]) bindings.toArray(new Binding[0]);
    }

    void addBinding(Binding b)
    {
// if this interface is a lhs-operand in the new binding,
// make sure to remove any existing binding where this
// is interface as a lhs-operand (if there is one)
        if(b.getLhsOperand() == this)
        {
            Binding[] binds = getBindings();
            for(int i = 0; i < binds.length; i++)
            {
                if(binds[i].getLhsOperand() == this)
                {
                    binds[i].getParent().removeBinding(binds[i]);
                    break;
                }
            }
        }

// add the new binding
        bindings.add(b);
    }

    void removeBinding(Binding bind)
    {
        bindings.remove(bind);
    }


    public Layout getLayout()
    {
        return layout;
    }

    private String name;
    private InterfaceType type;
    private int mode;
    private Component parent;
    private Collection bindings;
    private Layout layout = new Layout();

    // ARRAYS

    private String arrayProperty;

    public void setArrayProperty(String property)
    {

        if(property != arrayProperty && (property == null
                || !property.equals(arrayProperty)))
        {

            arrayProperty = property;
            if(property == null)
                parent.getType().getInterface(name).setArray(null);
            else
            {
                ArrayType arrayType = getType().getArrayType(property);
                parent.getType().getInterface(name).setArray(arrayType);
            }
            getParent().firePropertyChange("interfaceArray");
        }
    }

    public String getArrayProperty()
    {
        return arrayProperty;
    }

    public void removeReferences()
    {
        if(type != null) type.removeReference();
    }
}