package cz.cuni.dsrg.sofa.netbeans.visualdesigner.typemanager;

import cz.cuni.dsrg.sofa.netbeans.visualdesigner.types.*;
import org.openide.NotifyDescriptor;
import org.openide.TopManager;
import org.openide.actions.CopyAction;
import org.openide.actions.CutAction;
import org.openide.actions.DeleteAction;
import org.openide.actions.PasteAction;
import org.openide.explorer.ExplorerActions;
import org.openide.explorer.ExplorerManager;
import org.openide.explorer.ExplorerPanel;
import org.openide.explorer.view.BeanTreeView;
import org.openide.nodes.Node;
import org.openide.util.actions.SystemAction;

import javax.swing.*;
import java.awt.*;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.Map;

/**
 * This class is a JPanel implementing the type manager, a tree view
 * of the type information contained in the edited project.<p>
 *
 * The type information (frame, architecture, interface) is displayed by the
 * folders (repository, module, provide, architecture module). Different classes
 * of a type element is distinguished by several icons. If the type has multiple
 * versions, they are displayed as multiple subnodes of a parent node,
 * representing the defaut (i.e. most recent) version.<p>
 *
 * One or more types can be selected and the operations can be perfored using
 * the context menus or the toolbar.
 *
 * @author Michael Cifka
 */
public class TypeManager extends javax.swing.JPanel implements
        ExplorerManager.Provider, PropertyChangeListener
{

    public static String PROP_SELECTED_TYPE = "selectedType";

    public synchronized static TypeManager get(Repository repository)
    {
        TypeManager manager = (TypeManager) instances.get(repository);
        if(manager == null)
        {
            manager = new TypeManager(repository);
            instances.put(repository, manager);
        }
        return manager;
    }

    private TypeManager(Repository repository)
    {
        this.repository = repository;
        initComponents();

        actions = new Actions(this);
        JToolBar toolBar = actions.createToolBar();
        GridBagConstraints gridBagConstraints = new GridBagConstraints();
        gridBagConstraints.gridx = 0;
        gridBagConstraints.gridy = 0;
        gridBagConstraints.insets = new Insets(10, 10, 10, 10);
        add(toolBar, gridBagConstraints);


        explorerPanel = new ExplorerPanel();
        explorerPanel.add(new BeanTreeView());
        explorerActions = new ExplorerActions();
        getExplorerManager().setRootContext(new TypeNode(this, repository, true));
        getExplorerManager().addPropertyChangeListener(this);
        jPanel.add(explorerPanel);
    }

    // ACTIONS SUPPORT

    public void synchronize()
    {
        final java.awt.Dialog dialog = Synchronization.createDialog();

        // use a new thread to display the dialog; this thread will block on show()
        Thread t = new Thread()
        {
            public void run()
            {
                // wait until the dialog comes up
                while(!dialog.isVisible())
                    ; // do nothing
                try
                {
                    repository.synchronize();
                    dialog.hide();
                }
                catch (RepositoryException re)
                {
                    dialog.hide();
                    NotifyDescriptor d = new NotifyDescriptor.Message(
                            re.getMessage(), NotifyDescriptor.ERROR_MESSAGE);
                    TopManager.getDefault().notify(d);
                }
            }
        };
        t.start();
        dialog.show();
        refresh();
    }

    public String askForValidName(String title, Class clazz)
    {
        if(selectedType == null)
            return null;
        NotifyDescriptor.InputLine dlg =
                new NotifyDescriptor.InputLine("Name:", title);
        if(TopManager.getDefault().notify(dlg) != NotifyDescriptor.OK_OPTION)
            return null;
        String s = dlg.getInputText();
        AbstractType collision = selectedType.getDefaultChild(s);
        if(collision != null && (!collision.isCommited() || clazz == null
                || !clazz.isInstance(collision)))
        {
            NotifyDescriptor.Message msg =
                    new NotifyDescriptor.Message(
                            "Type " + s + " could not be created.", NotifyDescriptor.INFORMATION_MESSAGE);
            TopManager.getDefault().notify(msg);
            return null;
        }
        return s;
    }

    public void newFolder()
    {
        String s = askForValidName("New Module", null);
        if(s != null)
        {
            if(selectedType instanceof Module)
            {
                Module module = new Module(s);
                selectedType.addChild(module);
                refresh();
                setSelectedType(module);
            }
            else
            {
                AModule amodule = new AModule(s);
                selectedType.addChild(amodule);
                refresh();
                setSelectedType(amodule);
            }
        }
    }

    public void newInterface()
    {
        String s = askForValidName("New Interface", InterfaceType.class);
        if(s != null)
        {
            InterfaceType iface = new InterfaceType(s);
            selectedType.addChild(iface);
            refresh();
            setSelectedType(iface);
        }
    }

    public void newFrame()
    {
        String s = askForValidName("New Frame", FrameType.class);
        if(s != null)
        {
            FrameType frame = new FrameType(s);
            selectedType.addChild(frame);
            refresh();
            setSelectedType(frame);
        }
    }

    public void newProvider()
    {
        String s = askForValidName("New Provider", null);
        if(s != null)
        {
            Provider provider = new Provider(s);
            selectedType.addChild(provider);
            refresh();
            setSelectedType(provider);
        }
    }

    public void newArchitecture()
    {
        String s = askForValidName("New Architecture", ArchitectureType.class);
        if(s != null)
        {
            ArchitectureType arch = new ArchitectureType(s, null);
            selectedType.addChild(arch);
            refresh();
            setSelectedType(arch);
        }
    }

    public void duplicate()
    {
        if(selectedType != null)
        {
            AbstractType dup = selectedType.duplicate(selectedType.getParent());
            if(dup != null)
            {
                selectedType.getParent().addChild(dup);
                refresh();
                setSelectedType(dup);
            }
        }
    }

    public void cut()
    {
        CutAction cutAction = (CutAction) SystemAction.get(CutAction.class);
        cutAction.performAction();
    }

    public void copy()
    {
        CopyAction copyAction = (CopyAction) SystemAction.get(CopyAction.class);
        copyAction.performAction();
    }

    public void paste()
    {
        PasteAction pasteAction = (PasteAction) SystemAction.get(PasteAction.class);
        pasteAction.performAction();
    }

    public void delete()
    {
        DeleteAction deleteAction = (DeleteAction) SystemAction.get(DeleteAction.class);
        deleteAction.performAction();
    }

    public void refresh()
    {
        TypeNode root = (TypeNode) getExplorerManager().getRootContext();
        root.updateChildren();
    }

    public ExplorerManager getExplorerManager()
    {
        return explorerPanel.getExplorerManager();
    }

    public void setActions()
    {
        explorerActions.attach(getExplorerManager());
    }

    public void removeActions()
    {
        explorerActions.detach();
    }

    public TypeNode getSelectedNode()
    {
        return selectedNode;
    }

    public AbstractType getSelectedType()
    {
        return selectedType;
    }

    public void registerNode(TypeNode node)
    {
        nodes.put(node.getAssociatedType(), node);
    }

    public void unregisterNode(TypeNode node)
    {
        nodes.remove(node.getAssociatedType());
    }

    public void setSelectedType(AbstractType type)
    {

        ExplorerManager em = getExplorerManager();
        Node n = em.getRootContext();

        if(type == null)
        {
            em.setExploredContext(n, new Node[0]);
            actions.enable();
            return;
        }


        n = (Node) nodes.get(type);

        Node pn = n.getParentNode();
        if(pn != null)
            try
            {
                em.setExploredContext(pn, new Node[]{n});
            }
            catch (Exception ex)
            {
            }
        else
            em.setExploredContext(n, new Node[]{n});

        selectedType = type;
    }


    public void propertyChange(PropertyChangeEvent evt)
    {
        if(evt.getPropertyName().equals(ExplorerManager.PROP_SELECTED_NODES))
        {
            Node[] sel = (Node[]) evt.getNewValue();
            if(sel.length == 1)
            {
                selectedNode = (TypeNode) sel[0];
                selectedType = selectedNode.getAssociatedType();
                jLabel.setText(selectedType.getScopeNameVersion());
                this.firePropertyChange(PROP_SELECTED_TYPE, null, selectedType);
            }
            else
            {
                selectedNode = null;
                selectedType = null;
                jLabel.setText("[ no type selected ]");
                this.firePropertyChange(PROP_SELECTED_TYPE, null, null);
            }
            actions.enable();
        }
    }

    private String[] getPath(AbstractType type)
    {
        LinkedList list = new LinkedList();
        while(type.getParent() != null)
        {
            list.addFirst(type.getName());
            type = type.getParent();
        }
        return (String[]) list.toArray(new String[0]);
    }

    /** This method is called from within the constructor to
     * initialize the form.
     * WARNING: Do NOT modify this code. The content of this method is
     * always regenerated by the Form Editor.
     */
    private void initComponents()
    {//GEN-BEGIN:initComponents
        java.awt.GridBagConstraints gridBagConstraints;

        jPanel = new javax.swing.JPanel();
        jLabel = new javax.swing.JLabel();

        setLayout(new java.awt.GridBagLayout());

        jPanel.setLayout(new java.awt.BorderLayout());

        gridBagConstraints = new java.awt.GridBagConstraints();
        gridBagConstraints.gridx = 0;
        gridBagConstraints.gridy = 1;
        gridBagConstraints.fill = java.awt.GridBagConstraints.BOTH;
        gridBagConstraints.weightx = 1.0;
        gridBagConstraints.weighty = 1.0;
        gridBagConstraints.insets = new java.awt.Insets(0, 10, 10, 10);
        add(jPanel, gridBagConstraints);

        jLabel.setText("[ no type selected ]");
        jLabel.setFont(new java.awt.Font("Dialog", 0, 11));
        gridBagConstraints = new java.awt.GridBagConstraints();
        gridBagConstraints.gridx = 0;
        gridBagConstraints.gridy = 2;
        gridBagConstraints.fill = java.awt.GridBagConstraints.HORIZONTAL;
        gridBagConstraints.insets = new java.awt.Insets(0, 20, 10, 10);
        add(jLabel, gridBagConstraints);

    }//GEN-END:initComponents


    // Variables declaration - do not modify//GEN-BEGIN:variables
    private javax.swing.JLabel jLabel;
    private javax.swing.JPanel jPanel;
    // End of variables declaration//GEN-END:variables

    private ExplorerPanel explorerPanel;
    private ExplorerActions explorerActions;
    private Class requiredType;
    private AbstractType selectedType;
    private TypeNode selectedNode;
    private Repository repository;
    private Actions actions;
    private Map nodes = new HashMap();

    private static Map instances = new HashMap();
}

