package cz.cuni.dsrg.sofa.netbeans.visualdesigner.typemanager;

import cz.cuni.dsrg.sofa.netbeans.visualdesigner.types.*;
import org.openide.NotifyDescriptor;
import org.openide.TopManager;
import org.openide.actions.*;
import org.openide.nodes.AbstractNode;
import org.openide.nodes.Children;
import org.openide.nodes.Node;
import org.openide.util.actions.SystemAction;
import org.openide.util.datatransfer.ExTransferable;
import org.openide.util.datatransfer.NewType;
import org.openide.util.datatransfer.PasteType;

import java.awt.datatransfer.DataFlavor;
import java.awt.datatransfer.Transferable;
import java.io.IOException;
import java.util.Comparator;
import java.util.List;

/**
 * This node represents a type information (frame, architecture, interface) or
 * a type folder (repository, module, provide, architecture module). A large set
 * of operations to manipulate the represented component is provided.
 *
 * @author Michael Cifka
 */
public class TypeNode extends AbstractNode
{

    private TypeManager manager;

    /** Construct new node, either contents or module. */
    public TypeNode(TypeManager manager, AbstractType type, boolean defaultNode)
    {
        super(defaultNode ? new TypeNodeChildren() : Children.LEAF);
        this.manager = manager;
        this.type = type;
        this.defaultNode = defaultNode;
        updateDisplayName();

        if(type.isContainer())
        {
            AbstractType[] children = type.getDefaultChildren();
            for(int i = 0; i < children.length; i++)
            {
                if(!(children[i] instanceof ArrayType))
                {
                    TypeNode childNode = new TypeNode(manager, children[i], true);
                    getChildren().add(new Node[]{childNode});
                }
            }
        }
        else if(defaultNode)
        {
            AbstractType[] brothers = type.getParent().getChildrenForName(type.getName());
            for(int i = 0; i < brothers.length; i++)
            {
                TypeNode brotherNode = new TypeNode(manager, brothers[i], false);
                getChildren().add(new Node[]{brotherNode});
            }
        }
        updateIcon();
        manager.registerNode(this);
    }

    public void updateChildren()
    {

        updateIcon();
        updateName();
        updateDisplayName();

        Children children = getChildren();
        if(children == Children.LEAF) return;
        Node[] nodes = children.getNodes();

        // either children or versions
        if(type.isContainer())
        {
            AbstractType[] types = type.getDefaultChildren();
            updateChildrenNodes(nodes, types, true);
        }
        else
        {
            AbstractType p = type.getParent();
            if(p != null)
            {
                AbstractType[] types = p.getChildrenForName(type.getName());
                updateChildrenNodes(nodes, types, false);
            }
        }

        // update each child itself
        for(int i = 0; i < nodes.length; i++)
            ((TypeNode) nodes[i]).updateChildren();
    }

    private void updateChildrenNodes(Node[] nodes, AbstractType[] types, boolean defaultTypes)
    {

        Children children = getChildren();

        // REMOVE NODES
        // get list of types
        // go through all nodes, check if their type is in the list
        // if not, then remove the node


        for(int i = 0, j; i < nodes.length; i++)
        {
            AbstractType t = ((TypeNode) nodes[i]).getAssociatedType();
            for(j = 0; j < types.length; j++)
                if(t == types[j]) break;
            if(j == types.length)
            {   // surplus node -- remove
                children.remove(new Node[]{nodes[i]});
                manager.unregisterNode((TypeNode) nodes[i]);
            }
        }

        // ADD NODES
        // go through types, chcesk if there is a node for each of
        // them, if isn't then create one and add

        for(int i = 0, j; i < types.length; i++)
        {
            if(types[i] instanceof ArrayType) continue;
            for(j = 0; j < nodes.length; j++)
            {
                AbstractType t = ((TypeNode) nodes[j]).getAssociatedType();
                if(types[i] == t) break;
            }
            if(j == nodes.length) // node does not exist -- add
                children.add(new Node[]{new TypeNode(manager, types[i], defaultTypes)});
        }
    }

    public boolean canDestroy()
    {
        //return !type.isCommited() && type.isContainer() && type.getChildren().length == 0;
        return type.isRemoveable();
    }

    public boolean canRename()
    {
        return !type.isCommited();
    }

    public void destroy() throws IOException
    {
        AbstractType p = type.getParent();
        if(p != null)
        {
            p.removeChild(type);
            manager.refresh();
        }
    }

    // COPY, CUT & PASTE

    public static DataFlavor TYPE_FLAVOR = new DataFlavor(TypeTransferObject.class, "TypeTransfer Object");

    public class TypeTransferObject
    {
        public AbstractType type;
        public boolean remove;
    }

    public Transferable createTransferable(final AbstractType type, final boolean remove)
    {
        return new ExTransferable.Single(TYPE_FLAVOR)
        {
            public Object getData()
            {
                TypeTransferObject tto = new TypeTransferObject();
                tto.type = type;
                tto.remove = remove;
                return tto;
            }
        };
    }


    public boolean canCopy()
    {
        return !type.isContainer();
    }

    public boolean canCut()
    {
        return !type.isContainer() && !type.isCommited();
    }

    public Transferable clipboardCopy() throws IOException
    {
        return createTransferable(type, false);
    }

    public Transferable clipboardCut() throws IOException
    {
        return createTransferable(type, true);
    }

    protected void createPasteTypes(Transferable t, List ls)
    {
        TypeTransferObject tto = null;
        try
        {
            if(t.isDataFlavorSupported(TYPE_FLAVOR))
                tto = (TypeTransferObject) t.getTransferData(TYPE_FLAVOR);
        }
        catch (Exception e)
        {
        }
        ;
        if(tto != null
                && ((tto.type.getParent() instanceof Module || type instanceof Module)
                || (tto.type.getParent() instanceof AModule || type instanceof AModule)))
        {
            final TypeTransferObject obj = tto;
            ls.add(new PasteType()
            {
                public Transferable paste()
                {
                    if(obj.remove)
                    {
                        obj.type.getParent().removeChild(obj.type);
                        type.addChild(obj.type);
                    }
                    else
                    {
                        AbstractType duplicate = obj.type.duplicate(type);
                        type.addChild(duplicate);
                    }
                    manager.refresh();
                    return ExTransferable.EMPTY;
                }
            });
        }
        super.createPasteTypes(t, ls);
    }

    public void setName(String name)
    {
        if(getName().equals(name)) return;
        AbstractType collision = type.getParent().getDefaultChild(name);
        if(collision != null && (!collision.isCommited() || type.isContainer()
                || !type.getClass().isInstance(collision)))
        {
            NotifyDescriptor d = new NotifyDescriptor.Message(
                    "Cannot rename type " + getName() + " to " + name + ".",
                    NotifyDescriptor.INFORMATION_MESSAGE);
            TopManager.getDefault().notify(d);
            return;
        }
        type.setName(name);
        manager.refresh();
    }

    public void select()
    {
        setIconBase("org/netbeans/core/resources/defaultOpenFolder");
    }

    public void unselect()
    {
        setIconBase("org/netbeans/core/resources/defaultFolder");
    }

    public AbstractType getAssociatedType()
    {
        return type;
    }

    public NewType[] getNewTypes()
    {
        AbstractType type = contents ? this.type.getParent() : this.type;

        if(type instanceof Repository)
            return new NewType[]{newFolder, newInterface, newFrame, newProvider};
        else if(type instanceof Module)
            return new NewType[]{newFolder, newInterface, newFrame};
        else if(type instanceof Provider || type instanceof AModule)
            return new NewType[]{newFolder, newArchitecture};

        else
            return null;
    }

    protected SystemAction[] createActions()
    {

        if(type.isContainer())
            return new SystemAction[]{
                SystemAction.get(CutAction.class),
                SystemAction.get(CopyAction.class),
                SystemAction.get(PasteAction.class),
                null,
                SystemAction.get(DeleteAction.class),
                SystemAction.get(RenameAction.class),
                null,
                SystemAction.get(NewAction.class)
            };
        else
            return new SystemAction[]{
                SystemAction.get(CutAction.class),
                SystemAction.get(CopyAction.class),
                SystemAction.get(PasteAction.class),
                null,
                SystemAction.get(DeleteAction.class),
                SystemAction.get(RenameAction.class),
            };
    }


    private NewType newFolder = new NewType()
    {
        public String getName()
        {
            return "Module";
        }

        public void create()
        {
            manager.newFolder();
        }
    };

    private NewType newInterface = new NewType()
    {
        public String getName()
        {
            return "Interface";
        }

        public void create()
        {
            manager.newInterface();
        }
    };

    private NewType newFrame = new NewType()
    {
        public String getName()
        {
            return "Frame";
        }

        public void create()
        {
            manager.newFrame();
        }
    };

    private NewType newProvider = new NewType()
    {
        public String getName()
        {
            return "Provider";
        }

        public void create()
        {
            manager.newProvider();
        }
    };

    private NewType newArchitecture = new NewType()
    {
        public String getName()
        {
            return "Architecture";
        }

        public void create()
        {
            manager.newArchitecture();
        }
    };

    private void updateName()
    {
        super.setName((defaultNode || type.getVersion() == null)
                      ? type.getName() : type.getVersion());
    }

    private void updateDisplayName()
    {
        String displayName;
        if(defaultNode)
            displayName = type.getName();
        else if(type.getVersion() == null)
            displayName = "uncommited";
        else
            displayName = type.getVersion();
        setDisplayName(displayName);
    }

    private void updateIcon()
    {
        String base = "cz/cuni/dsrg/sofa/netbeans/resources/visualdesigner/icons/";
        if(type instanceof Repository)
            base = base + "repository";
        else if(!defaultNode)
            base = base + "version";
        else if(type instanceof Provider)
            base = base + "provider";
        else if(type instanceof Module)
            base = base + "folder";
        else if(type instanceof AModule)
            base = base + "folder";
        else if(type instanceof InterfaceType)
            base = base + "interface";
        else if(type instanceof FrameType)
            base = base + "frame";
        else if(type instanceof ArchitectureType) base = base + "architecture";
        if(type.isCommited() && !(type instanceof Repository)) base = base + "Commited";
        setIconBase(base);
    }

    private AbstractType type;
    private boolean defaultNode;
    private boolean contents;
    private TypeNode peer;

    private static class TypeNodeChildren extends Children.SortedArray implements Comparator
    {
        public TypeNodeChildren()
        {
            setComparator(this);
        }

        public int compare(Object objA, Object objB)
        {
            AbstractType a = ((TypeNode) objA).getAssociatedType();
            AbstractType b = ((TypeNode) objB).getAssociatedType();
            if(a.isContainer() && !b.isContainer()) return -1;
            if(!a.isContainer() && b.isContainer()) return 1;
            if(a instanceof Module && b instanceof Provider) return -1;
            if(a instanceof Provider && b instanceof Module) return 1;
            return a.getName().compareToIgnoreCase(b.getName());
        }
    }
}