package cz.cuni.dsrg.sofa.netbeans.visualdesigner.types;

import SOFA.SOFAnode.Made.TIR.*;

import java.rmi.RemoteException;
import java.util.*;

/**
 * This class extends the {@link AbstractType} to hold, import, and export
 * the type information representing an architecture.<p>
 *
 * Reference to the frame type implemented by the architecture is managed
 * as well as the references to the instances and binding, encapsulated
 * using the {@link ArchitectureType.Instance} and
 * {@link ArchitectureType.Binding} classes.<p>
 *
 * This class can be persisted using the Java Object Serialization.
 *
 * @author Michael Cifka
 */
public class ArchitectureType extends AbstractType implements java.io.Serializable
{

    static final long serialVersionUID = -5976589989729844283L;

    public class Instance implements java.io.Serializable
    {
        public Instance(String name, String typeFullName, String arrayFullName)
        {
            this.name = name;
            this.typeFullName = typeFullName;
            this.arrayFullName = arrayFullName;
        }

        public Instance(String name, FrameType type)
        {
            this.name = name;
            this.type = type;
        }

        public String getName()
        {
            return name;
        }

        void setName(String name)
        {
            this.name = name;
        }

        public FrameType getType()
        {
            if(type == null)
                type = (FrameType) root.findObject(typeFullName);
            return type;
        }

        public void setType(FrameType type)
        {
            this.type = type;
        }

        public ArrayType getArray()
        {
            if(array == null)
                array = (ArrayType) root.findObject(arrayFullName);
            return array;
        }

        public void setArray(ArrayType array)
        {
            arrayFullName = null;
            this.array = array;
        }

        private String name;
        private FrameType type;
        private ArrayType array;
        private String typeFullName;
        private String arrayFullName;
    }

    public class Binding implements java.io.Serializable
    {
        public static final int BIND = 0;
        public static final int SUBSUME = 1;
        public static final int DELEGATE = 2;

        private ArchitectureType.Instance lhsInstance, rhsInstance;
        private FrameType.Interface lhsInterface, rhsInterface;

        public Binding(String lhsInstanceName, String lhsInterfaceName, String rhsInstanceName, String rhsInterfaceName)
        {
            this.lhsInstance = (lhsInstanceName == null) ? null : getInstance(lhsInstanceName);
            this.lhsInterface = (lhsInstance == null) ? getImplementedFrame().getInterface(lhsInterfaceName) : lhsInstance.getType().getInterface(lhsInterfaceName);
            this.rhsInstance = (rhsInstanceName == null) ? null : getInstance(rhsInstanceName);
            this.rhsInterface = (rhsInstance == null) ? getImplementedFrame().getInterface(rhsInterfaceName) : rhsInstance.getType().getInterface(rhsInterfaceName);
            mode = (lhsInstance == null) ? DELEGATE : ((rhsInstance == null) ? SUBSUME : BIND);

        }

        public int getMode()
        {
            return mode;
        }

        public String getLhsInstance()
        {
            return (lhsInstance == null) ? null : lhsInstance.getName();
        }

        public String getLhsInterface()
        {
            return lhsInterface.getName();
        }

        public String getRhsInstance()
        {
            return (rhsInstance == null) ? null : rhsInstance.getName();
        }

        public String getRhsInterface()
        {
            return rhsInterface.getName();
        }

        public boolean equals(Object obj)
        {
            if(obj instanceof Binding)
            {
                Binding other = (Binding) obj;
                return equals(getLhsInstance(), other.getLhsInstance()) && equals(getLhsInterface(), other.getLhsInterface())
                        && equals(getRhsInstance(), other.getRhsInstance()) && equals(getRhsInterface(), other.getRhsInterface());
            }
            else
                return false;
        }

        private boolean equals(String s1, String s2)
        {
            return s1 == null && s2 == null || s1 != null && s1.equals(s2);
        }

        private int mode;
    }

    ArchitectureType()
    {
        container = false;
        instances = new LinkedList();
        bindings = new LinkedList();
        frame = null;
        instancesIndex = new HashMap();
        properties = new HashMap();
    }

    public ArchitectureType(String name, FrameType frame)
    {
        this();
        this.name = name;
        this.frame = frame;
    }

    private ArchitectureType(String name, ArchitectureType other)
    {
        this(name, other.getImplementedFrame());
        Instance[] insts = other.getInstances();
        for(int i = 0; i < insts.length; i++)
            addInstance(insts[i].getName(), insts[i].getType());
        Binding[] binds = other.getBindings();
        for(int i = 0; i < binds.length; i++)
            addBinding(binds[i].getLhsInstance(), binds[i].getLhsInterface(),
                       binds[i].getRhsInstance(), binds[i].getRhsInterface());
    }

    public AbstractType duplicate(AbstractType parent)
    {
        return new ArchitectureType(parent.getAvailableName(name, ArchitectureType.class), this);
    }

    public FrameType getImplementedFrame()
    {
        if(frame == null)
            frame = (FrameType) root.findObject(frameFullName);
        return frame;
    }

    public void setImplementedFrame(FrameType frame)
    {
        if(getImplementedFrame() == null)
            this.frame = frame;
    }

    public Instance[] getInstances()
    {
        return (Instance[]) instances.toArray(new Instance[0]);
    }

    public Instance getInstance(String name)
    {
        return (Instance) instancesIndex.get(name);
    }

    public void addInstance(String name, FrameType type)
    {
        Instance inst = new Instance(name, type);
        instancesIndex.put(name, inst);
        instances.add(inst);
    }

    public void removeInstance(String name)
    {
        Object obj = instancesIndex.remove(name);
        instances.remove(obj);
    }

    public void renameInstance(String oldName, String newName)
    {
        Instance inst = (Instance) instancesIndex.remove(oldName);
        if(inst != null)
        {
            inst.setName(newName);
            instancesIndex.put(newName, inst);
        }
    }

    public Binding[] getBindings()
    {
        return (Binding[]) bindings.toArray(new Binding[0]);
    }

    public void addBinding(String lhsInstance, String lhsInterface, String rhsInstance, String rhsInterface)
    {

        Binding bind = new Binding(lhsInstance, lhsInterface, rhsInstance, rhsInterface);
        bindings.add(bind);
    }

    public void removeBinding(String lhsInstance, String lhsInterface, String rhsInstance, String rhsInterface)
    {

        Binding toRemove = new Binding(lhsInstance, lhsInterface, rhsInstance, rhsInterface);
        for(Iterator i = bindings.iterator(); i.hasNext();)
            if(i.next().equals(toRemove))
            {
                i.remove();
                break;
            }
    }


    public Property[] getProperties()
    {
        return (Property[]) properties.values().toArray(new Property[0]);
    }

    public Property getProperty()
    {
        return (Property) properties.get(name);
    }

    public void addProperty(String name, String type)
    {
        properties.put(name, new Property(name, type));
    }

    public void removeProperty(String name)
    {
        properties.remove(name);
    }

    protected void loadCommitedObjects(TIRObject o)
    {
        if(loaded) return;
        loaded = true;
        try
        {
            Container cont = (Container) o;
            frameFullName = getFullName(((ArchitectureDef) cont).frame());
            Contained objs[] = cont.contents(null);
            for(int i = 0; i < objs.length; i++)
            {
                if(objs[i] instanceof InstDef)
                {
                    String arrayFullName = null;
                    CDLType type = ((InstDef) objs[i]).type();
                    if(type instanceof TypedefDef)
                    {
                        arrayFullName = getFullName((TypedefDef) type);
                        type = ((TypedefDef) type).original_type();
                        if(!(type instanceof ArrayDef)) continue;
                        type = ((ArrayDef) type).element_type();
                    }
                    if(type instanceof FrameDef)
                    {
                        String name = objs[i].get_identification().name();
                        Instance inst = new Instance(name, getFullName((FrameDef) type), arrayFullName);
                        instancesIndex.put(name, inst);
                        instances.add(inst);
                    }
                }
                else if(objs[i] instanceof BindDef)
                {
                    BindDef bindDef = (BindDef) objs[i];
                    int mode = bindDef.mode().value();
                    String lhsInstanceName = null, lhsInterfaceName = null,
                            rhsInstanceName = null, rhsInterfaceName = null;
                    try
                    {

                        // exception will occur when loading array binds which are not supported

                        switch(mode)
                        {
                            case Binding.BIND:
                                BindOperSub lhsOp = (BindOperSub) bindDef.lhsop();
                                BindOperSub rhsOp = (BindOperSub) bindDef.rhsop();
                                BindType lhsInstance = lhsOp.subs()[0];
                                lhsInstanceName = (lhsInstance instanceof BindTypeNormal)
                                        ? ((BindTypeNormal) lhsInstance).name() : ((BindTypeArray) lhsInstance).name();
                                BindType lhsInterface = lhsOp.elements()[0];
                                lhsInterfaceName = (lhsInterface instanceof BindTypeNormal)
                                        ? ((BindTypeNormal) lhsInterface).name() : ((BindTypeArray) lhsInterface).name();
                                BindType rhsInstance = rhsOp.subs()[0];
                                rhsInstanceName = (rhsInstance instanceof BindTypeNormal)
                                        ? ((BindTypeNormal) rhsInstance).name() : ((BindTypeArray) rhsInstance).name();
                                BindType rhsInterface = rhsOp.elements()[0];
                                rhsInterfaceName = (rhsInterface instanceof BindTypeNormal)
                                        ? ((BindTypeNormal) rhsInterface).name() : ((BindTypeArray) rhsInterface).name();
                                break;
                            case Binding.SUBSUME:
                                lhsOp = (BindOperSub) bindDef.lhsop();
                                BindOperProvReq rhsIfaceOp = (BindOperProvReq) bindDef.rhsop();
                                lhsInstance = lhsOp.subs()[0];
                                lhsInstanceName = (lhsInstance instanceof BindTypeNormal)
                                        ? ((BindTypeNormal) lhsInstance).name() : ((BindTypeArray) lhsInstance).name();
                                lhsInterface = lhsOp.elements()[0];
                                lhsInterfaceName = (lhsInterface instanceof BindTypeNormal)
                                        ? ((BindTypeNormal) lhsInterface).name() : ((BindTypeArray) lhsInterface).name();
                                rhsInstanceName = null;
                                rhsInterface = rhsIfaceOp.elements()[0];
                                rhsInterfaceName = (rhsInterface instanceof BindTypeNormal)
                                        ? ((BindTypeNormal) rhsInterface).name() : ((BindTypeArray) rhsInterface).name();
                                break;
                            case Binding.DELEGATE:
                                BindOperProvReq lhsIfaceOp = (BindOperProvReq) bindDef.lhsop();
                                rhsOp = (BindOperSub) bindDef.rhsop();
                                lhsInstanceName = null;
                                lhsInterface = lhsIfaceOp.elements()[0];
                                lhsInterfaceName = (lhsInterface instanceof BindTypeNormal)
                                        ? ((BindTypeNormal) lhsInterface).name() : ((BindTypeArray) lhsInterface).name();
                                rhsInstance = rhsOp.subs()[0];
                                rhsInstanceName = (rhsInstance instanceof BindTypeNormal)
                                        ? ((BindTypeNormal) rhsInstance).name() : ((BindTypeArray) rhsInstance).name();
                                rhsInterface = rhsOp.elements()[0];
                                rhsInterfaceName = (rhsInterface instanceof BindTypeNormal)
                                        ? ((BindTypeNormal) rhsInterface).name() : ((BindTypeArray) rhsInterface).name();
                                break;
                        }
                        Binding bind = new Binding(lhsInstanceName, lhsInterfaceName, rhsInstanceName, rhsInterfaceName);
                        bindings.add(bind);
                    }
                    catch (ClassCastException e)
                    {
                    }
                }
                else if(objs[i] instanceof PropertyDef)
                {
                    String name = objs[i].get_identification().name();
                    CDLType type = ((PropertyDef) objs[i]).type();
                    if(type instanceof PrimitiveDef)
                    {
                        int kind = ((PrimitiveDef) type).kind().value();
                        if(kind == PrimitiveKind.pk_string || kind == PrimitiveKind.pk_wstring)
                            addProperty(name, "string");
                        else if(kind == PrimitiveKind.pk_long || kind == PrimitiveKind.pk_longlong || kind == PrimitiveKind.pk_short
                                || kind == PrimitiveKind.pk_ulong || kind == PrimitiveKind.pk_ulonglong || kind == PrimitiveKind.pk_ushort)
                            addProperty(name, "integer");
                    }
                }
            }
        }
        catch (Exception e)
        {
            System.out.println("Exception: " + e.getMessage());
            e.printStackTrace();
        }
    }

    private FrameType frame;
    private String frameFullName;
    private Collection instances;
    private Collection bindings;
    private Map instancesIndex;
    protected static int counter;
    private boolean loaded;
    private Map properties;

    private String getFullName(Contained obj) throws RemoteException
    {
        Identification id = obj.get_identification();
        String absoluteName = id.absolute_name().name();
        return absoluteName.substring(absoluteName.indexOf("::"))
                + " ver. " + id.version();
    }

    public String getProviderName()
    {
        AbstractType p = parent;
        while(!(p instanceof Provider))
            p = p.getParent();
        return p.getName();
    }

    public String writeAsCDL()
    {

        if(commited) return "";


        String s = "\narchitecture " + getProviderName() + " " + getRelativeName(null)
                + "\nimplements " +
                 (getImplementedFrame().getName().equalsIgnoreCase("SYSTEM")? "::SOFA::libs::Application" : getImplementedFrame().getRelativeName(null))
                 + " ";

        Instance[] insts = getInstances();
        Property[] prop = getProperties();
        if(insts.length == 0 && prop.length == 0)
        {
            s += "primitive;\n";
            return s;
        }
        s += "{\n";
        String indentString = indentString(1);

        // write properties
        for(int i = 0; i < prop.length; i++)
            s += indentString + "property " + prop[i].getType()
                    + " " + prop[i].getName() + ";\n";
        if(prop.length > 0 && insts.length > 0)
            s += "\n";

        String t = "";
        for(int i = 0; i < insts.length; i++)
            if(insts[i].getArray() != null)
                t += indentString + "inst " + insts[i].getArray().getRelativeName(null)
                        + " " + insts[i].getName() + ";\n";
            else
                t += indentString + "inst " + insts[i].getType().getRelativeName(null)
                        + " " + insts[i].getName() + ";\n";

        s += (t.equals("")) ? "" : t + "\n";

        Binding[] binds = getBindings();
        String[] keywords = new String[]{"bind", "subsume", "delegate"};
        for(int i = 0; i < binds.length; i++)
        {
            int mode = binds[i].getMode();
            FrameType currentFrame;
            s += indentString + keywords[mode] + " ";
            if(mode != Binding.DELEGATE)
            {
                s += binds[i].getLhsInstance();
                Instance lhsInstance = getInstance(binds[i].getLhsInstance());
                if(lhsInstance.getArray() != null)
                    s += "[<1..$" + lhsInstance.getArray().getProperty() + ">]";
                s += ":";
                currentFrame = lhsInstance.getType();
            }
            else
                currentFrame = getImplementedFrame();
            s += binds[i].getLhsInterface();
            FrameType.Interface iface = currentFrame.getInterface(binds[i].getLhsInterface());
            if(/*iface != null &&*/ iface.getArray() != null)
                s += "[<1..$" + iface.getArray().getProperty() + ">]";
            s += " to ";
            if(mode != Binding.SUBSUME)
            {
                s += binds[i].getRhsInstance();
                Instance rhsInstance = getInstance(binds[i].getRhsInstance());
                if(rhsInstance.getArray() != null)
                    s += "[<1..$" + rhsInstance.getArray().getProperty() + ">]";
                s += ":";
                currentFrame = rhsInstance.getType();
            }
            else
                currentFrame = getImplementedFrame();
            s += binds[i].getRhsInterface();
            iface = currentFrame.getInterface(binds[i].getRhsInterface());
            if(/*iface != null &&*/ iface.getArray() != null)
                s += "[<1..$" + iface.getArray().getProperty() + ">]";
            s += ";\n";
        }
        s += "};\n";
        return s;
    }
}